# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Protocol, TypeVar

from no_vtf.image import Image, ImageDataTypes
from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture, contravariant=True)


class TextureDecoder(Protocol[_TextureTypeVar]):
    @abstractmethod
    def __call__(self, texture: _TextureTypeVar) -> Image[ImageDataTypes]:
        ...



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import dataclasses

from dataclasses import dataclass
from typing import Optional, cast

from no_vtf.parsers.generated.vtf import Vtf as VtfParser
from no_vtf.texture import Texture


@dataclass(frozen=True, kw_only=True)
class VtfTexture(Texture):
    mipmap_level: int
    frame_index: Optional[int]
    face_index: Optional[int]
    slice_index: Optional[int]

    image: VtfParser.Image = dataclasses.field(repr=False, hash=False)

    def get_width(self) -> int:
        return cast(int, self.image.logical_width)

    def get_height(self) -> int:
        return cast(int, self.image.logical_height)

    def get_data(self) -> bytes:
        return cast(bytes, self.image.image_data)



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from no_vtf.deferrable import Deferrable, from_deferrable
from no_vtf.io import Io


class BytesIo(Io[bytes]):
    def write(self, path: pathlib.Path, data: Deferrable[bytes]) -> None:
        data = from_deferrable(data)

        with open(path, "wb") as file:
            file.write(data)

    def readback(self, path: pathlib.Path, data: Deferrable[bytes]) -> None:
        data = from_deferrable(data)

        with open(path, "rb") as file:
            read_data = file.read()
            if data != read_data:
                raise RuntimeError("Data differs from what is in the file")



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import multiprocessing

from no_vtf.main import main

if __name__ == "__main__":
    multiprocessing.freeze_support()

    main()



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import contextlib
import inspect
import pathlib
import sys

from typing import Optional, Sequence

import alive_progress
import click

import no_vtf

from no_vtf.files.vtf import VtfFile
from no_vtf.filesystem import DirectoryFilesList
from no_vtf.image import DynamicRange
from no_vtf.pipeline import Pipeline
from no_vtf.pipeline_runner import PipelineRunner, Task
from no_vtf.pipeline_runners.parallel_runner import ParallelRunner
from no_vtf.pipeline_runners.sequential_runner import SequentialRunner
from no_vtf.textures.vtf import VtfTexture


def _show_credits(ctx: click.Context, param: click.Parameter, value: bool) -> None:
    if not value or ctx.resilient_parsing:
        return

    credits = """
    no_vtf - Valve Texture Format Converter
    Copyright (C) 2022 b5327157

    https://sr.ht/~b5327157/no_vtf/
    https://pypi.org/project/no-vtf/

    This program is free software: you can redistribute it and/or modify it under
    the terms of the GNU Lesser General Public License as published by the Free
    Software Foundation, either version 3 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT ANY
    WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
    PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program. If not, see <https://www.gnu.org/licenses/>.
    """

    click.echo(inspect.cleandoc(credits))
    ctx.exit()


@click.command(name="no_vtf", no_args_is_help=True)
@click.argument(
    "paths",
    metavar="PATH...",
    type=click.Path(path_type=pathlib.Path, exists=True),
    required=True,
    nargs=-1,
)
@click.option(
    "--output-dir",
    "-o",
    "output_directory",
    help="Output directory",
    type=click.Path(
        path_type=pathlib.Path, exists=True, file_okay=False, dir_okay=True, writable=True
    ),
)
@click.option(
    "--ldr-format", "-l", help="LDR output format", show_default=True, type=str, default="tiff"
)
@click.option(
    "--hdr-format", "-h", help="HDR output format", show_default=True, type=str, default="exr"
)
@click.option(
    "--dynamic-range",
    "-d",
    help="Override LDR/HDR auto-detection",
    type=click.Choice(["ldr", "hdr"], case_sensitive=False),
)
@click.option(
    "--mipmaps",
    "-m",
    help="Extract all mipmaps",
    type=bool,
    is_flag=True,
)
@click.option(
    "--separate-channels",
    "-s",
    help="Output the RGB/L and A channels separately",
    type=bool,
    is_flag=True,
)
@click.option(
    "--overbright-factor",
    "-O",
    help="Multiplicative factor used for decoding compressed HDR textures",
    show_default=True,
    type=float,
    default=16,
)
@click.option(
    "--compress/--no-compress", help="Control lossless compression", type=bool, default=None
)
@click.option("write", "--always-write/--no-write", help="Write images", type=bool, default=None)
@click.option(
    "readback", "--readback/--no-readback", help="Readback images", type=bool, default=False
)
@click.option(
    "--num-workers",
    help="Number of workers for parallel conversion",
    metavar="INTEGER",
    type=click.IntRange(min=1),
)
@click.option(
    "--no-progress",
    help="Do not show the progress bar",
    type=bool,
    is_flag=True,
)
@click.version_option(version=no_vtf.__version__, message="%(version)s")
@click.option(
    "--credits",
    help="Show the credits and exit.",
    type=bool,
    is_flag=True,
    expose_value=False,
    is_eager=True,
    callback=_show_credits,
)
def main(
    *,
    paths: Sequence[pathlib.Path],
    output_directory: Optional[pathlib.Path],
    ldr_format: str,
    hdr_format: str,
    dynamic_range: Optional[DynamicRange],
    mipmaps: bool,
    separate_channels: bool,
    overbright_factor: float,
    compress: Optional[bool],
    write: Optional[bool],
    readback: bool,
    num_workers: Optional[int],
    no_progress: bool,
) -> None:
    """
    Convert Valve Texture Format files into standard image files.

    PATH can be either file, or directory (in which case it is recursively searched for .vtf files).
    Multiple paths may be provided.

    If the output directory is not specified, images are output into the source directories.
    Otherwise, directory tree for any found files will be reconstructed in the chosen directory.

    Output LDR/HDR format is selected by its common file name extension.
    Special formats:
    "raw" to write the high resolution image data as-is;
    "skip" to skip the write step entirely.

    For supported formats, compression is controlled when saving the image.
    Lossless compression is enabled by default. Lossy compression is not used.

    The BGRA8888 format can store both LDR and compressed HDR images.
    The specific type is either auto-detected by looking at the input file name
    (roughly, if it contains "hdr" near the end), or can be set manually.

    Only the highest-resolution mipmap is extracted by default.
    Alternatively, all mipmaps of the high-resolution image can be extracted.

    The RGB/L and A channels are packed into one file by default.
    When output separately, resulting file names will be suffixed with "_rgb", "_l" or "_a".

    By default, image files are only written if they do not exist already.
    Alternatively, they can be overwritten, or writing can be disabled entirely.

    Images can be also read back to verify they have been written properly.
    Readback will error if data to be written do not match what is in the file.

    Worker is spawned for each logical core to run the conversion in parallel.
    Number of workers can be overridden. If set to 1, conversion is sequential.

    Exit status: Zero if all went successfully, non-zero if there was an error.
    Upon a recoverable error, conversion will proceed with the next file.
    """

    _global_config()

    directory_files_list = DirectoryFilesList.from_paths(paths)
    if directory_files_list.has_unresolved_files():
        _resolve_files(directory_files_list, not no_progress)

    file_factory = VtfFile.make_factory(
        mipmaps=mipmaps, dynamic_range=dynamic_range, overbright_factor=overbright_factor
    )

    pipeline = Pipeline(
        file_factory=file_factory,
        ldr_format=ldr_format,
        hdr_format=hdr_format,
        separate_channels=separate_channels,
        compress=compress,
        write=write,
        readback=readback,
    )

    pipeline_runner: PipelineRunner
    if num_workers is None or num_workers > 1:
        pipeline_runner = ParallelRunner(max_workers=num_workers)
    else:
        pipeline_runner = SequentialRunner()

    tasks = _get_tasks(directory_files_list, output_directory)
    exit_status = _process_tasks(pipeline_runner, tasks, pipeline, not no_progress)
    sys.exit(exit_status)


def _global_config() -> None:
    alive_progress.config_handler.set_global(spinner=None, theme="classic", enrich_print=False)


def _resolve_files(directory_files_list: DirectoryFilesList, show_progress: bool) -> None:
    progress_bar_manager = alive_progress.alive_bar(receipt=False) if show_progress else None
    with progress_bar_manager or contextlib.nullcontext() as progress_bar:
        for file in directory_files_list.search_in_directories("*.vtf"):
            if progress_bar:
                progress_bar.text = file.name
                progress_bar()


def _get_tasks(
    directory_files_list: DirectoryFilesList,
    output_directory: Optional[pathlib.Path],
) -> Sequence[Task]:
    tasks: list[Task] = []
    for directory_files in directory_files_list:
        for file in directory_files.files:
            file_root_output_directory = output_directory or directory_files.directory
            file_relative_to_directory = file.relative_to(directory_files.directory)
            file_relative_directory = file_relative_to_directory.parent
            file_output_directory = file_root_output_directory / file_relative_directory
            tasks.append(Task(input_file=file, output_directory=file_output_directory))
    return tasks


def _process_tasks(
    pipeline_runner: PipelineRunner,
    tasks: Sequence[Task],
    pipeline: Pipeline[VtfTexture],
    show_progress: bool,
) -> int:
    exit_status = 0

    num_files = len(tasks)
    progress_bar_manager = alive_progress.alive_bar(num_files) if show_progress else None
    with progress_bar_manager or contextlib.nullcontext() as progress_bar:
        for result in pipeline_runner(pipeline, tasks):
            if not result.exception:
                if progress_bar:
                    progress_bar()
                    progress_bar.text = result.task.input_file.name
            else:
                exit_status = 1

                message = f'Error while processing "{result.task.input_file}": {result.exception}'
                if result.exception.__cause__:
                    message += f" ({result.exception.__cause__})"
                click.echo(message, file=sys.stderr)

    return exit_status



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from abc import abstractmethod
from typing import Protocol, Sequence, TypeVar

from no_vtf.image import Image, ImageDataTypes
from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture)


class File(Protocol[_TextureTypeVar]):
    @abstractmethod
    def extract(self) -> Sequence[_TextureTypeVar]:
        ...

    @abstractmethod
    def name(self, texture: _TextureTypeVar) -> str:
        ...

    @abstractmethod
    def decode(self, texture: _TextureTypeVar) -> Image[ImageDataTypes]:
        ...


class FileFactory(Protocol[_TextureTypeVar]):
    @abstractmethod
    def __call__(self, path: pathlib.Path) -> File[_TextureTypeVar]:
        ...



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Vtf(KaitaiStruct):
    """
    .. seealso::
       Source - https://developer.valvesoftware.com/wiki/Valve_Texture_Format
    
    
    .. seealso::
       Source - https://github.com/NeilJed/VTFLib
    """

    class ImageFormat(Enum):
        rgba8888 = 0
        abgr8888 = 1
        rgb888 = 2
        bgr888 = 3
        rgb565 = 4
        i8 = 5
        ia88 = 6
        p8 = 7
        a8 = 8
        rgb888_bluescreen = 9
        bgr888_bluescreen = 10
        argb8888 = 11
        bgra8888 = 12
        dxt1 = 13
        dxt3 = 14
        dxt5 = 15
        bgrx8888 = 16
        bgr565 = 17
        bgrx5551 = 18
        bgra4444 = 19
        dxt1_onebitalpha = 20
        bgra5551 = 21
        uv88 = 22
        uvwq8888 = 23
        rgba16161616f = 24
        rgba16161616 = 25
        uvlx8888 = 26
        none = 4294967295

    class ResourceTag(Enum):
        low_res_image = 65536
        particle_sheet = 1048576
        high_res_image = 3145728
        crc = 4411971
        key_values_data = 4937284
        lod_control = 5001028
        extended_flags = 5526351
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._read()

    def _read(self):
        self.header = Vtf.Header(self._io, self, self._root)

    class Body(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            if  ((self._root.header.base.version.minor < 3) and (self._root.header.v7_0.low_res_image_format != Vtf.ImageFormat.none) and (self._root.header.v7_0.low_res_image_width != 0) and (self._root.header.v7_0.low_res_image_height != 0)) :
                self.low_res_image = Vtf.Image(self._root.header.v7_0.low_res_image_width, self._root.header.v7_0.low_res_image_height, self._root.header.v7_0.low_res_image_format, self._io, self, self._root)

            if  ((self._root.header.base.version.minor < 3) and (self._root.header.v7_0.high_res_image_format != Vtf.ImageFormat.none)) :
                self.high_res_image = Vtf.HighResImage(self._io, self, self._root)

            if self._root.header.base.version.minor >= 3:
                self.resources = [None] * (self._root.header.logical.num_resources)
                for i in range(self._root.header.logical.num_resources):
                    self.resources[i] = Vtf.Resource(self._io, self, self._root)




    class Image(KaitaiStruct):
        def __init__(self, logical_width, logical_height, image_format, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.logical_width = logical_width
            self.logical_height = logical_height
            self.image_format = image_format
            self._read()

        def _read(self):
            self.image_data = self._io.read_bytes(((self.physical_width * self.physical_height) * self.bpp) // 8)

        @property
        def image_format_to_bpp(self):
            if hasattr(self, '_m_image_format_to_bpp'):
                return self._m_image_format_to_bpp if hasattr(self, '_m_image_format_to_bpp') else None

            self._m_image_format_to_bpp = Vtf.ImageFormatToBpp(self.image_format, self._io, self, self._root)
            return self._m_image_format_to_bpp if hasattr(self, '_m_image_format_to_bpp') else None

        @property
        def physical_height(self):
            if hasattr(self, '_m_physical_height'):
                return self._m_physical_height if hasattr(self, '_m_physical_height') else None

            self._m_physical_height = (self.logical_height if not (self.is_compressed) else (self.logical_height if self.logical_height >= 4 else 4))
            return self._m_physical_height if hasattr(self, '_m_physical_height') else None

        @property
        def is_compressed(self):
            if hasattr(self, '_m_is_compressed'):
                return self._m_is_compressed if hasattr(self, '_m_is_compressed') else None

            self._m_is_compressed = self.image_format_to_is_compressed.is_compressed_container.value
            return self._m_is_compressed if hasattr(self, '_m_is_compressed') else None

        @property
        def bpp(self):
            if hasattr(self, '_m_bpp'):
                return self._m_bpp if hasattr(self, '_m_bpp') else None

            self._m_bpp = self.image_format_to_bpp.bpp_container.value
            return self._m_bpp if hasattr(self, '_m_bpp') else None

        @property
        def physical_width(self):
            if hasattr(self, '_m_physical_width'):
                return self._m_physical_width if hasattr(self, '_m_physical_width') else None

            self._m_physical_width = (self.logical_width if not (self.is_compressed) else (self.logical_width if self.logical_width >= 4 else 4))
            return self._m_physical_width if hasattr(self, '_m_physical_width') else None

        @property
        def image_format_to_is_compressed(self):
            if hasattr(self, '_m_image_format_to_is_compressed'):
                return self._m_image_format_to_is_compressed if hasattr(self, '_m_image_format_to_is_compressed') else None

            self._m_image_format_to_is_compressed = Vtf.ImageFormatToIsCompressed(self.image_format, self._io, self, self._root)
            return self._m_image_format_to_is_compressed if hasattr(self, '_m_image_format_to_is_compressed') else None


    class ImageFace(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.image_slices = [None] * (self._parent.num_slices)
            for i in range(self._parent.num_slices):
                self.image_slices[i] = Vtf.Image(self._parent.width, self._parent.height, self._parent.image_format, self._io, self._parent, self._root)



    class ImageFormatToIsCompressed(KaitaiStruct):
        def __init__(self, image_format, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.image_format = image_format
            self._read()

        def _read(self):
            pass

        @property
        def is_compressed_container(self):
            if hasattr(self, '_m_is_compressed_container'):
                return self._m_is_compressed_container if hasattr(self, '_m_is_compressed_container') else None

            _on = self.image_format
            if _on == Vtf.ImageFormat.dxt5:
                self._m_is_compressed_container = Vtf.B1Container(True, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.dxt1_onebitalpha:
                self._m_is_compressed_container = Vtf.B1Container(True, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.dxt1:
                self._m_is_compressed_container = Vtf.B1Container(True, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.dxt3:
                self._m_is_compressed_container = Vtf.B1Container(True, self._io, self, self._root)
            else:
                self._m_is_compressed_container = Vtf.B1Container(False, self._io, self, self._root)
            return self._m_is_compressed_container if hasattr(self, '_m_is_compressed_container') else None


    class HeaderV72(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.num_slices = self._io.read_u2le()
            if not self.num_slices >= 1:
                raise kaitaistruct.ValidationLessThanError(1, self.num_slices, self._io, u"/types/header_v7_2/seq/0")


    class HeaderBase(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.magic = self._io.read_bytes(4)
            if not self.magic == b"\x56\x54\x46\x00":
                raise kaitaistruct.ValidationNotEqualError(b"\x56\x54\x46\x00", self.magic, self._io, u"/types/header_base/seq/0")
            self.version = Vtf.Version(self._io, self, self._root)
            self.header_size = self._io.read_u4le()


    class HeaderFlags(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def onebitalpha(self):
            if hasattr(self, '_m_onebitalpha'):
                return self._m_onebitalpha if hasattr(self, '_m_onebitalpha') else None

            self._m_onebitalpha = (True if (self._root.header.v7_0.flags & 4096) != 0 else False)
            return self._m_onebitalpha if hasattr(self, '_m_onebitalpha') else None

        @property
        def eightbitalpha(self):
            if hasattr(self, '_m_eightbitalpha'):
                return self._m_eightbitalpha if hasattr(self, '_m_eightbitalpha') else None

            self._m_eightbitalpha = (True if (self._root.header.v7_0.flags & 8192) != 0 else False)
            return self._m_eightbitalpha if hasattr(self, '_m_eightbitalpha') else None

        @property
        def envmap(self):
            if hasattr(self, '_m_envmap'):
                return self._m_envmap if hasattr(self, '_m_envmap') else None

            self._m_envmap = (True if (self._root.header.v7_0.flags & 16384) != 0 else False)
            return self._m_envmap if hasattr(self, '_m_envmap') else None


    class Version(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.major = self._io.read_u4le()
            if not self.major >= 7:
                raise kaitaistruct.ValidationLessThanError(7, self.major, self._io, u"/types/version/seq/0")
            if not self.major <= 7:
                raise kaitaistruct.ValidationGreaterThanError(7, self.major, self._io, u"/types/version/seq/0")
            self.minor = self._io.read_u4le()
            if not self.minor >= 0:
                raise kaitaistruct.ValidationLessThanError(0, self.minor, self._io, u"/types/version/seq/1")
            if not self.minor <= 5:
                raise kaitaistruct.ValidationGreaterThanError(5, self.minor, self._io, u"/types/version/seq/1")


    class HighResImage(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.image_mipmaps = [None] * (self._root.header.v7_0.num_mipmaps)
            for i in range(self._root.header.v7_0.num_mipmaps):
                self.image_mipmaps[i] = Vtf.ImageMipmap(i, self._io, self, self._root)



    class ImageFrame(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.image_faces = [None] * (self._root.header.logical.num_faces)
            for i in range(self._root.header.logical.num_faces):
                self.image_faces[i] = Vtf.ImageFace(self._io, self._parent, self._root)



    class Resource(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.tag = KaitaiStream.resolve_enum(Vtf.ResourceTag, self._io.read_bits_int_be(24))
            self._io.align_to_byte()
            self.flags = self._io.read_u1()
            if  ((False) or (self.tag == Vtf.ResourceTag.low_res_image) or (self.tag == Vtf.ResourceTag.high_res_image) or (self.tag == Vtf.ResourceTag.particle_sheet) or (self.tag == Vtf.ResourceTag.key_values_data)) :
                self.ofs_resource = self._io.read_u4le()

            if  ((True) and (self.tag != Vtf.ResourceTag.low_res_image) and (self.tag != Vtf.ResourceTag.high_res_image) and (self.tag != Vtf.ResourceTag.particle_sheet) and (self.tag != Vtf.ResourceTag.key_values_data)) :
                self._unnamed3 = self._io.read_u4le()


        @property
        def high_res_image(self):
            if hasattr(self, '_m_high_res_image'):
                return self._m_high_res_image if hasattr(self, '_m_high_res_image') else None

            if  ((self.tag == Vtf.ResourceTag.high_res_image) and (self._root.header.v7_0.high_res_image_format != Vtf.ImageFormat.none)) :
                _pos = self._io.pos()
                self._io.seek(self.ofs_resource)
                self._m_high_res_image = Vtf.HighResImage(self._io, self, self._root)
                self._io.seek(_pos)

            return self._m_high_res_image if hasattr(self, '_m_high_res_image') else None

        @property
        def low_res_image(self):
            if hasattr(self, '_m_low_res_image'):
                return self._m_low_res_image if hasattr(self, '_m_low_res_image') else None

            if  ((self.tag == Vtf.ResourceTag.low_res_image) and (self._root.header.v7_0.low_res_image_format != Vtf.ImageFormat.none)) :
                _pos = self._io.pos()
                self._io.seek(self.ofs_resource)
                self._m_low_res_image = Vtf.Image(self._root.header.v7_0.low_res_image_width, self._root.header.v7_0.low_res_image_height, self._root.header.v7_0.low_res_image_format, self._io, self, self._root)
                self._io.seek(_pos)

            return self._m_low_res_image if hasattr(self, '_m_low_res_image') else None


    class Header(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.base = Vtf.HeaderBase(self._io, self, self._root)
            self.v7_0 = Vtf.HeaderV70(self._io, self, self._root)
            if self.base.version.minor >= 2:
                self.v7_2 = Vtf.HeaderV72(self._io, self, self._root)

            if self.base.version.minor >= 3:
                self.v7_3 = Vtf.HeaderV73(self._io, self, self._root)

            self.logical = Vtf.HeaderLogical(self._io, self, self._root)


    class B1Container(KaitaiStruct):
        def __init__(self, value, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.value = value
            self._read()

        def _read(self):
            pass


    class HeaderV73(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.reserved2 = self._io.read_bytes(3)
            self.num_resources = self._io.read_u4le()
            self.reserved3 = self._io.read_bytes(8)


    class HeaderLogical(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            pass

        @property
        def flags(self):
            if hasattr(self, '_m_flags'):
                return self._m_flags if hasattr(self, '_m_flags') else None

            self._m_flags = Vtf.HeaderFlags(self._io, self, self._root)
            return self._m_flags if hasattr(self, '_m_flags') else None

        @property
        def num_faces(self):
            if hasattr(self, '_m_num_faces'):
                return self._m_num_faces if hasattr(self, '_m_num_faces') else None

            self._m_num_faces = (1 if not (self.flags.envmap) else (6 if  ((self._root.header.base.version.minor < 1) or (self._root.header.base.version.minor > 4))  else (6 if self._root.header.v7_0.first_frame == 255 else 7)))
            return self._m_num_faces if hasattr(self, '_m_num_faces') else None

        @property
        def num_slices(self):
            if hasattr(self, '_m_num_slices'):
                return self._m_num_slices if hasattr(self, '_m_num_slices') else None

            self._m_num_slices = (self._root.header.v7_2.num_slices if self._root.header.base.version.minor >= 2 else 1)
            return self._m_num_slices if hasattr(self, '_m_num_slices') else None

        @property
        def num_resources(self):
            if hasattr(self, '_m_num_resources'):
                return self._m_num_resources if hasattr(self, '_m_num_resources') else None

            self._m_num_resources = (self._root.header.v7_3.num_resources if self._root.header.base.version.minor >= 3 else 0)
            return self._m_num_resources if hasattr(self, '_m_num_resources') else None


    class HeaderV70(KaitaiStruct):
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._read()

        def _read(self):
            self.width = self._io.read_u2le()
            if not self.width >= 1:
                raise kaitaistruct.ValidationLessThanError(1, self.width, self._io, u"/types/header_v7_0/seq/0")
            self.height = self._io.read_u2le()
            if not self.height >= 1:
                raise kaitaistruct.ValidationLessThanError(1, self.height, self._io, u"/types/header_v7_0/seq/1")
            self.flags = self._io.read_u4le()
            self.num_frames = self._io.read_u2le()
            self.first_frame = self._io.read_u2le()
            self.reserved0 = self._io.read_bytes(4)
            self.reflectivity = [None] * (3)
            for i in range(3):
                self.reflectivity[i] = self._io.read_f4le()

            self.reserved1 = self._io.read_bytes(4)
            self.bumpmap_scale = self._io.read_f4le()
            self.high_res_image_format = KaitaiStream.resolve_enum(Vtf.ImageFormat, self._io.read_u4le())
            self.num_mipmaps = self._io.read_u1()
            self.low_res_image_format = KaitaiStream.resolve_enum(Vtf.ImageFormat, self._io.read_u4le())
            self.low_res_image_width = self._io.read_u1()
            self.low_res_image_height = self._io.read_u1()


    class U4Container(KaitaiStruct):
        def __init__(self, value, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.value = value
            self._read()

        def _read(self):
            pass


    class ImageFormatToBpp(KaitaiStruct):
        def __init__(self, image_format, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.image_format = image_format
            self._read()

        def _read(self):
            pass

        @property
        def bpp_container(self):
            if hasattr(self, '_m_bpp_container'):
                return self._m_bpp_container if hasattr(self, '_m_bpp_container') else None

            _on = self.image_format
            if _on == Vtf.ImageFormat.argb8888:
                self._m_bpp_container = Vtf.U4Container(32, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgra8888:
                self._m_bpp_container = Vtf.U4Container(32, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.abgr8888:
                self._m_bpp_container = Vtf.U4Container(32, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.rgb888_bluescreen:
                self._m_bpp_container = Vtf.U4Container(24, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.rgba16161616:
                self._m_bpp_container = Vtf.U4Container(64, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.dxt5:
                self._m_bpp_container = Vtf.U4Container(8, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.a8:
                self._m_bpp_container = Vtf.U4Container(8, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.ia88:
                self._m_bpp_container = Vtf.U4Container(16, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgr565:
                self._m_bpp_container = Vtf.U4Container(16, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.uvlx8888:
                self._m_bpp_container = Vtf.U4Container(32, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgr888:
                self._m_bpp_container = Vtf.U4Container(24, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.dxt1_onebitalpha:
                self._m_bpp_container = Vtf.U4Container(4, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.p8:
                self._m_bpp_container = Vtf.U4Container(8, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.rgb888:
                self._m_bpp_container = Vtf.U4Container(24, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.rgba8888:
                self._m_bpp_container = Vtf.U4Container(32, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.i8:
                self._m_bpp_container = Vtf.U4Container(8, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgrx5551:
                self._m_bpp_container = Vtf.U4Container(16, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.uvwq8888:
                self._m_bpp_container = Vtf.U4Container(32, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgr888_bluescreen:
                self._m_bpp_container = Vtf.U4Container(24, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgra5551:
                self._m_bpp_container = Vtf.U4Container(16, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgrx8888:
                self._m_bpp_container = Vtf.U4Container(32, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.bgra4444:
                self._m_bpp_container = Vtf.U4Container(16, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.dxt1:
                self._m_bpp_container = Vtf.U4Container(4, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.dxt3:
                self._m_bpp_container = Vtf.U4Container(8, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.rgb565:
                self._m_bpp_container = Vtf.U4Container(16, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.uv88:
                self._m_bpp_container = Vtf.U4Container(16, self._io, self, self._root)
            elif _on == Vtf.ImageFormat.rgba16161616f:
                self._m_bpp_container = Vtf.U4Container(64, self._io, self, self._root)
            return self._m_bpp_container if hasattr(self, '_m_bpp_container') else None


    class ImageMipmap(KaitaiStruct):
        def __init__(self, mipmap_index, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.mipmap_index = mipmap_index
            self._read()

        def _read(self):
            self.image_frames = [None] * (self._root.header.v7_0.num_frames)
            for i in range(self._root.header.v7_0.num_frames):
                self.image_frames[i] = Vtf.ImageFrame(self._io, self, self._root)


        @property
        def height(self):
            if hasattr(self, '_m_height'):
                return self._m_height if hasattr(self, '_m_height') else None

            self._m_height = ((self._root.header.v7_0.height >> self.bit_shift) if (self._root.header.v7_0.height >> self.bit_shift) >= 1 else 1)
            return self._m_height if hasattr(self, '_m_height') else None

        @property
        def image_format(self):
            if hasattr(self, '_m_image_format'):
                return self._m_image_format if hasattr(self, '_m_image_format') else None

            self._m_image_format = self._root.header.v7_0.high_res_image_format
            return self._m_image_format if hasattr(self, '_m_image_format') else None

        @property
        def num_slices(self):
            if hasattr(self, '_m_num_slices'):
                return self._m_num_slices if hasattr(self, '_m_num_slices') else None

            self._m_num_slices = ((self._root.header.logical.num_slices >> self.bit_shift) if (self._root.header.logical.num_slices >> self.bit_shift) >= 1 else 1)
            return self._m_num_slices if hasattr(self, '_m_num_slices') else None

        @property
        def width(self):
            if hasattr(self, '_m_width'):
                return self._m_width if hasattr(self, '_m_width') else None

            self._m_width = ((self._root.header.v7_0.width >> self.bit_shift) if (self._root.header.v7_0.width >> self.bit_shift) >= 1 else 1)
            return self._m_width if hasattr(self, '_m_width') else None

        @property
        def bit_shift(self):
            if hasattr(self, '_m_bit_shift'):
                return self._m_bit_shift if hasattr(self, '_m_bit_shift') else None

            self._m_bit_shift = ((self._root.header.v7_0.num_mipmaps - self.mipmap_index) - 1)
            return self._m_bit_shift if hasattr(self, '_m_bit_shift') else None


    @property
    def body(self):
        if hasattr(self, '_m_body'):
            return self._m_body if hasattr(self, '_m_body') else None

        _pos = self._io.pos()
        self._io.seek((self.header.base.header_size - (self.header.logical.num_resources * 8)))
        self._m_body = Vtf.Body(self._io, self, self._root)
        self._io.seek(_pos)
        return self._m_body if hasattr(self, '_m_body') else None





# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools

from dataclasses import dataclass
from typing import Optional

from no_vtf.decoder.image import ImageDecoder
from no_vtf.decoder.texture import TextureDecoder
from no_vtf.decoders.image.generic import (
    decode_a_uint8,
    decode_abgr_uint8,
    decode_argb_uint8,
    decode_bgr_uint8,
    decode_bgra_uint4_le,
    decode_bgra_uint8,
    decode_dxt1_rgb,
    decode_dxt1_rgba,
    decode_dxt3,
    decode_dxt5,
    decode_l_uint8,
    decode_la_uint8,
    decode_rgb_uint8,
    decode_rgba_float16_le,
    decode_rgba_uint8,
    decode_uv_uint8,
)
from no_vtf.decoders.image.vtf import (
    decode_bgr_uint8_bluescreen,
    decode_bgra_uint8_hdr,
    decode_rgb_uint8_bluescreen,
    decode_rgba_uint16_le_hdr,
)
from no_vtf.image import DynamicRange, Image, ImageDataTypes
from no_vtf.parsers.generated.vtf import Vtf as VtfParser
from no_vtf.textures.vtf import VtfTexture


@dataclass(frozen=True, kw_only=True)
class VtfDecoder(TextureDecoder[VtfTexture]):
    dynamic_range: DynamicRange
    overbright_factor: Optional[float] = None

    def __call__(self, texture: VtfTexture) -> Image[ImageDataTypes]:
        encoded_image = texture.image.image_data
        logical_width = texture.image.logical_width
        logical_height = texture.image.logical_height

        decoder = self._get_decoder(texture)
        decoded_image = decoder(encoded_image, logical_width, logical_height)
        return decoded_image

    def _get_decoder(self, texture: VtfTexture) -> ImageDecoder[ImageDataTypes]:
        image_format = texture.image.image_format

        decoder: Optional[ImageDecoder[ImageDataTypes]] = None
        match (image_format, self.dynamic_range):
            case VtfParser.ImageFormat.rgba8888, _:
                decoder = decode_rgba_uint8
            case VtfParser.ImageFormat.abgr8888, _:
                decoder = decode_abgr_uint8
            case VtfParser.ImageFormat.rgb888, _:
                decoder = decode_rgb_uint8
            case VtfParser.ImageFormat.bgr888, _:
                decoder = decode_bgr_uint8
            case VtfParser.ImageFormat.i8, _:
                decoder = decode_l_uint8
            case VtfParser.ImageFormat.ia88, _:
                decoder = decode_la_uint8
            case VtfParser.ImageFormat.a8, _:
                decoder = decode_a_uint8
            case VtfParser.ImageFormat.rgb888_bluescreen, _:
                decoder = decode_rgb_uint8_bluescreen
            case VtfParser.ImageFormat.bgr888_bluescreen, _:
                decoder = decode_bgr_uint8_bluescreen
            case VtfParser.ImageFormat.argb8888, _:
                # VTFLib/VTFEdit, Gimp VTF Plugin, and possibly others, decode this format
                # differently because of mismatched channels (verified against VTF2TGA).
                decoder = decode_argb_uint8
            case VtfParser.ImageFormat.bgra8888, "ldr":
                decoder = decode_bgra_uint8
            case VtfParser.ImageFormat.bgra8888, "hdr":
                decoder = functools.partial(
                    decode_bgra_uint8_hdr, overbright_factor=self.overbright_factor
                )
            case VtfParser.ImageFormat.dxt1, _:
                decoder = decode_dxt1_rgb
            case VtfParser.ImageFormat.dxt3, _:
                decoder = decode_dxt3
            case VtfParser.ImageFormat.dxt5, _:
                decoder = decode_dxt5
            case VtfParser.ImageFormat.bgra4444, _:
                decoder = decode_bgra_uint4_le
            case VtfParser.ImageFormat.dxt1_onebitalpha, _:
                decoder = decode_dxt1_rgba
            case VtfParser.ImageFormat.uv88, _:
                decoder = decode_uv_uint8
            case VtfParser.ImageFormat.rgba16161616f, _:
                decoder = decode_rgba_float16_le
            case VtfParser.ImageFormat.rgba16161616, _:
                decoder = decode_rgba_uint16_le_hdr

        if not decoder:
            raise RuntimeError(f"Unsupported Valve texture format: {image_format.name}")

        return decoder



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from abc import abstractmethod
from typing import Protocol, TypeVar

from no_vtf.deferrable import Deferrable

_IoTypeVar = TypeVar("_IoTypeVar", contravariant=True)


class Io(Protocol[_IoTypeVar]):
    @abstractmethod
    def write(self, path: pathlib.Path, data: Deferrable[_IoTypeVar]) -> None:
        ...

    @abstractmethod
    def readback(self, path: pathlib.Path, data: Deferrable[_IoTypeVar]) -> None:
        ...



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools
import multiprocessing
import signal

from dataclasses import dataclass
from typing import Iterable, Optional, Sequence, TypeVar

from no_vtf.pipeline import Pipeline
from no_vtf.pipeline_runner import PipelineRunner, Result, Task
from no_vtf.pipeline_runners.sequential_runner import SequentialRunner
from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture)


@dataclass(frozen=True, kw_only=True)
class ParallelRunner(PipelineRunner):
    max_workers: Optional[int] = None

    def __call__(
        self, pipeline: Pipeline[_TextureTypeVar], tasks: Sequence[Task]
    ) -> Iterable[Result]:
        ignore_sigint = functools.partial(signal.signal, signal.SIGINT, signal.SIG_IGN)
        with multiprocessing.Pool(self.max_workers, initializer=ignore_sigint) as pool:
            process_with_pipeline = functools.partial(SequentialRunner.process, pipeline)
            for result in pool.imap_unordered(process_with_pipeline, tasks):
                yield result



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Sequence

from no_vtf.filter import Filter
from no_vtf.textures.vtf import VtfTexture


class HighestResolutionMipmapFilter(Filter[VtfTexture]):
    def __call__(self, textures: Sequence[VtfTexture]) -> Sequence[VtfTexture]:
        highest_resolution_textures = [texture for texture in textures if texture.mipmap_level == 0]
        return highest_resolution_textures



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import pathlib

from abc import abstractmethod
from dataclasses import dataclass
from typing import Iterable, Optional, Protocol, Sequence, TypeVar

from no_vtf.pipeline import Pipeline
from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture)


class PipelineRunner(Protocol):
    @abstractmethod
    def __call__(
        self, pipeline: Pipeline[_TextureTypeVar], tasks: Sequence[Task]
    ) -> Iterable[Result]:
        ...


@dataclass(frozen=True, kw_only=True)
class Task:
    input_file: pathlib.Path
    output_directory: pathlib.Path


@dataclass(frozen=True, kw_only=True)
class Result:
    task: Task
    exception: Optional[Exception]



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from dataclasses import dataclass
from typing import BinaryIO, Optional, Sequence, cast

import kaitaistruct

from no_vtf.extractor import Extractor
from no_vtf.parsers.generated.vtf import Vtf as VtfParser
from no_vtf.textures.vtf import VtfTexture


@dataclass(frozen=True, kw_only=True)
class VtfExtractor(Extractor[VtfTexture]):
    def __call__(self, io: BinaryIO) -> Sequence[VtfTexture]:
        try:
            parser = VtfParser.from_io(io)
        except kaitaistruct.KaitaiStructError as exception:
            raise RuntimeError(f"Parser error ({exception})")

        high_res_image = self._high_res_image_7_0(parser) or self._high_res_image_7_3(parser)
        if not high_res_image:
            return ()

        textures = self._textures_from_high_res_image(parser, high_res_image)
        return textures

    def _high_res_image_7_0(self, parser: VtfParser) -> Optional[VtfParser.HighResImage]:
        return getattr(parser.body, "high_res_image", None)

    def _high_res_image_7_3(self, parser: VtfParser) -> Optional[VtfParser.HighResImage]:
        resources = getattr(parser.body, "resources", None)
        if not resources:
            return None

        for resource in resources:
            high_res_image: Optional[VtfParser.HighResImage] = getattr(
                resource, "high_res_image", None
            )
            if high_res_image:
                return high_res_image

        return None

    def _textures_from_high_res_image(
        self, parser: VtfParser, high_res_image: VtfParser.HighResImage
    ) -> Sequence[VtfTexture]:
        has_multiple_frames = parser.header.v7_0.num_frames > 1
        is_cubemap = parser.header.logical.flags.envmap
        has_multiple_slices = parser.header.logical.num_slices > 1

        textures: list[VtfTexture] = []
        for mipmap in high_res_image.image_mipmaps:
            mipmap = cast(VtfParser.ImageMipmap, mipmap)
            mipmap_level = parser.header.v7_0.num_mipmaps - mipmap.mipmap_index - 1

            frame_index: Optional[int]
            for frame_index, frame in enumerate(mipmap.image_frames):
                frame_index = frame_index if has_multiple_frames else None
                frame = cast(VtfParser.ImageFrame, frame)

                face_index: Optional[int]
                for face_index, face in enumerate(frame.image_faces):
                    face_index = face_index if is_cubemap else None
                    face = cast(VtfParser.ImageFace, face)

                    slice_index: Optional[int]
                    for slice_index, slice in enumerate(face.image_slices):
                        slice_index = slice_index if has_multiple_slices else None
                        slice = cast(VtfParser.Image, slice)

                        texture = VtfTexture(
                            mipmap_level=mipmap_level,
                            frame_index=frame_index,
                            face_index=face_index,
                            slice_index=slice_index,
                            image=slice,
                        )
                        textures.append(texture)
        return textures



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Protocol, Sequence, TypeVar

from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture)


class Filter(Protocol[_TextureTypeVar]):
    @abstractmethod
    def __call__(self, textures: Sequence[_TextureTypeVar]) -> Sequence[_TextureTypeVar]:
        ...


def apply_filters(
    filters: Sequence[Filter[_TextureTypeVar]], textures: Sequence[_TextureTypeVar]
) -> Sequence[_TextureTypeVar]:
    for filter in filters:
        textures = filter(textures)
    return textures



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import functools
import pathlib
import re

from typing import ClassVar, Optional, Sequence

from no_vtf.decoders.texture.vtf import VtfDecoder
from no_vtf.extractors.vtf import VtfExtractor
from no_vtf.file import File, FileFactory
from no_vtf.filter import apply_filters
from no_vtf.filters.vtf import HighestResolutionMipmapFilter
from no_vtf.image import DynamicRange, Image, ImageDataTypes
from no_vtf.namers.vtf import VTF2TGALikeNamer
from no_vtf.textures.vtf import VtfTexture


class VtfFile(File[VtfTexture]):
    @staticmethod
    def make_factory(
        *,
        mipmaps: bool = False,
        dynamic_range: Optional[DynamicRange] = None,
        overbright_factor: Optional[float] = None,
    ) -> FileFactory[VtfTexture]:
        return functools.partial(
            VtfFile,
            mipmaps=mipmaps,
            dynamic_range=dynamic_range,
            overbright_factor=overbright_factor,
        )

    def __init__(
        self,
        path: pathlib.Path,
        *,
        mipmaps: bool = False,
        dynamic_range: Optional[DynamicRange] = None,
        overbright_factor: Optional[float] = None,
    ):
        if dynamic_range is None:
            if self._hdr_file_name_pattern.search(path.name) is not None:
                dynamic_range = "hdr"
            else:
                dynamic_range = "ldr"

        self._path = path
        self._mipmaps = mipmaps
        self._dynamic_range = dynamic_range
        self._overbright_factor = overbright_factor

        self._textures: Optional[Sequence[VtfTexture]] = None

    _hdr_file_name_pattern: ClassVar[re.Pattern[str]] = re.compile(
        r"[_\.] \d*? hdr .*? \.vtf $", re.ASCII | re.IGNORECASE | re.VERBOSE
    )

    def extract(self) -> Sequence[VtfTexture]:
        if self._textures is not None:
            return list(self._textures)

        with open(self._path, "rb") as io:
            extractor = VtfExtractor()
            self._textures = extractor(io)

            filters = []
            if not self._mipmaps:
                filters.append(HighestResolutionMipmapFilter())
            self._textures = apply_filters(filters, self._textures)

        return list(self._textures)

    def name(self, texture: VtfTexture) -> str:
        name_stem = self._path.stem

        textures = self.extract()
        last_mipmap_level = max(texture.mipmap_level for texture in textures)

        namer = VTF2TGALikeNamer(include_mipmap_level=(last_mipmap_level != 0))
        return namer(name_stem, texture)

    def decode(self, texture: VtfTexture) -> Image[ImageDataTypes]:
        decoder = VtfDecoder(
            dynamic_range=self._dynamic_range, overbright_factor=self._overbright_factor
        )
        return decoder(texture)



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import TYPE_CHECKING, Any, Optional, Sequence, Type, TypeVar

import numpy as np
import numpy.typing as npt

if TYPE_CHECKING:
    ByteOrder = np._ByteOrder
else:
    ByteOrder = Any

_ScalarTypeVar = TypeVar("_ScalarTypeVar", bound=np.generic, covariant=True)


def image_ndarray(
    image: bytes,
    width: int,
    height: int,
    channel_order: Sequence[int],
    scalar_type: Type[_ScalarTypeVar],
    byte_order: Optional[ByteOrder] = None,
) -> npt.NDArray[_ScalarTypeVar]:
    num_channels = len(channel_order)
    shape = (height, width, num_channels)

    dtype = np.dtype(scalar_type)
    if byte_order is not None:
        dtype = dtype.newbyteorder(byte_order)

    ndarray: npt.NDArray[_ScalarTypeVar] = np.ndarray(shape=shape, dtype=dtype, buffer=image).copy()
    ndarray = ndarray[..., channel_order]
    return ndarray



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import pathlib

from dataclasses import dataclass
from typing import Generic, Literal, Optional, TypeVar

from no_vtf.channel_separator import ChannelSeparator
from no_vtf.file import FileFactory
from no_vtf.filesystem import FileAccessor
from no_vtf.image import Image, ImageDataTypes
from no_vtf.io import Io
from no_vtf.ios.bytes_io import BytesIo
from no_vtf.ios.imageio import Imageio
from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture)
_IoTypeVar = TypeVar("_IoTypeVar", contravariant=True)


@dataclass(frozen=True, kw_only=True)
class Pipeline(Generic[_TextureTypeVar]):
    FORMAT_RAW: Literal["raw"] = "raw"
    FORMAT_SKIP: Literal["skip"] = "skip"

    file_factory: FileFactory[_TextureTypeVar]

    ldr_format: str
    hdr_format: str

    separate_channels: bool = False

    compress: Optional[bool] = None

    write: Optional[bool] = None
    readback: bool = False

    def __post_init__(self) -> None:
        Imageio.request_freeimage()

    def __call__(self, input_file: pathlib.Path, output_directory: pathlib.Path) -> None:
        texture_file = self.file_factory(input_file)
        textures = texture_file.extract()
        for texture in textures:
            image = texture_file.decode(texture)
            dynamic_range = image.get_dynamic_range()

            format = self.hdr_format if dynamic_range == "hdr" else self.ldr_format
            if _compare_formats(format, self.FORMAT_SKIP):
                continue

            name_stem = texture_file.name(texture)
            output_name = name_stem + "." + format
            output_path = output_directory / output_name

            if _compare_formats(format, self.FORMAT_RAW):
                bytes_io = BytesIo()
                data = texture.get_data()
                self._do_io(bytes_io, output_path, data)
            else:
                imageio = Imageio(format=format, compress=self.compress)
                if not self.separate_channels:
                    self._do_io(imageio, output_path, image)
                else:
                    self._separate_channels(imageio, output_path, image)

    def _separate_channels(
        self, imageio: Imageio, path: pathlib.Path, image: Image[ImageDataTypes]
    ) -> None:
        channel_separator = ChannelSeparator()
        for image_separated in channel_separator(image):
            new_stem = path.stem + "_" + image_separated.channels
            new_path = path.with_stem(new_stem)
            self._do_io(imageio, new_path, image_separated)

    def _do_io(self, io: Io[_IoTypeVar], path: pathlib.Path, data: _IoTypeVar) -> None:
        if self.write is not False:
            self._write(io, path, data)

        if self.readback:
            io.readback(path, data)

    def _write(self, io: Io[_IoTypeVar], path: pathlib.Path, data: _IoTypeVar) -> None:
        skip_existing = self.write is None
        file_accessor = FileAccessor(
            skip_existing=skip_existing,
            mkdir_missing_parents=True,
        )

        if file_accessor(path):
            io.write(path, data)


def _compare_formats(a: str, b: str) -> bool:
    return a.lower() == b.lower()



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from dataclasses import dataclass
from typing import Literal, Optional, TypeVar

import numpy as np
import numpy.typing as npt

from no_vtf.image import ImageDataTypes
from no_vtf.postprocessor import Postprocessor

_ImageDataTypeVar = TypeVar(
    "_ImageDataTypeVar",
    bound=ImageDataTypes,
    covariant=True,
)

FloatingPointNumBits = Literal[16, 32, 64]


@dataclass(frozen=True, kw_only=True)
class FloatingPointPrecisionModifier(Postprocessor[npt.NDArray[_ImageDataTypeVar]]):
    min: Optional[FloatingPointNumBits] = None
    forced: Optional[FloatingPointNumBits] = None
    max: Optional[FloatingPointNumBits] = None

    def __post_init__(self) -> None:
        precisions = [
            precision for precision in (self.min, self.forced, self.max) if precision is not None
        ]
        precisions_sorted = sorted(precisions)
        if precisions != precisions_sorted:
            raise RuntimeError(
                f"Unordered precisions: {self.min = }, {self.forced = }, {self.max = }"
            )

    def __call__(self, data: npt.NDArray[_ImageDataTypeVar]) -> npt.NDArray[_ImageDataTypeVar]:
        if not np.issubdtype(data.dtype, np.floating):
            return data

        if self.forced is not None:
            return data.astype(f"float{self.forced}")

        fp_bits = data.dtype.itemsize * 8
        if self.min is not None and fp_bits < self.min:
            return data.astype(f"float{self.min}")
        if self.max is not None and fp_bits > self.max:
            return data.astype(f"float{self.max}")

        return data



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Protocol, TypeVar

from no_vtf.image import Image, ImageDataTypes

_ImageDataTypeVar = TypeVar(
    "_ImageDataTypeVar",
    bound=ImageDataTypes,
    covariant=True,
)


class ImageDecoder(Protocol[_ImageDataTypeVar]):
    @abstractmethod
    def __call__(
        self, encoded_image: bytes, logical_width: int, logical_height: int, /
    ) -> Image[_ImageDataTypeVar]:
        ...



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import BinaryIO, Protocol, Sequence, TypeVar

from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture, covariant=True)


class Extractor(Protocol[_TextureTypeVar]):
    @abstractmethod
    def __call__(self, io: BinaryIO) -> Sequence[_TextureTypeVar]:
        ...



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Optional

import numpy as np
import numpy.typing as npt

from no_vtf.decoders.image.generic import (
    decode_bgr_uint8,
    decode_bgra_uint8,
    decode_rgb_uint8,
    decode_rgba_uint16_le,
)
from no_vtf.deferrable import Deferred
from no_vtf.image import Image


def decode_rgb_uint8_bluescreen(encoded_image: bytes, width: int, height: int) -> Image[np.uint8]:
    image = decode_rgb_uint8(encoded_image, width, height)
    image = _decode_bluescreen(image)
    return image


def decode_bgr_uint8_bluescreen(encoded_image: bytes, width: int, height: int) -> Image[np.uint8]:
    image = decode_bgr_uint8(encoded_image, width, height)
    image = _decode_bluescreen(image)
    return image


def _decode_bluescreen(image: Image[np.uint8]) -> Image[np.uint8]:
    assert image.channels == "rgb"

    def thunk() -> npt.NDArray[np.uint8]:
        rgb_uint8 = image.data()

        is_opaque: npt.NDArray[np.bool_] = rgb_uint8 != (0, 0, 255)
        is_opaque = is_opaque.any(axis=2)
        is_opaque = is_opaque[..., np.newaxis]

        rgb_uint8 *= is_opaque
        a_uint8: npt.NDArray[np.uint8] = np.multiply(is_opaque, 255, dtype=np.uint8)

        rgba_uint8: npt.NDArray[np.uint8] = np.dstack((rgb_uint8, a_uint8))
        return rgba_uint8

    return Image(data=Deferred(thunk), dtype=np.uint8, channels="rgba")


def decode_bgra_uint8_hdr(
    encoded_image: bytes, width: int, height: int, overbright_factor: Optional[float]
) -> Image[np.float32]:
    if overbright_factor is None:
        overbright_factor = 16

    def thunk() -> npt.NDArray[np.float32]:
        rgba_uint8 = decode_bgra_uint8(encoded_image, width, height).data()

        rgba_float32: npt.NDArray[np.float32] = rgba_uint8.astype(np.float32) / 255.0
        rgba_float32[:, :, [0, 1, 2]] *= rgba_float32[:, :, [3]] * overbright_factor

        rgb_float32: npt.NDArray[np.float32] = rgba_float32[..., :3]
        return rgb_float32

    return Image(data=Deferred(thunk), dtype=np.float32, channels="rgb")


def decode_rgba_uint16_le_hdr(encoded_image: bytes, width: int, height: int) -> Image[np.float32]:
    def thunk() -> npt.NDArray[np.float32]:
        rgba_uint16 = decode_rgba_uint16_le(encoded_image, width, height).data()
        # convert 4.12 fixed point stored as integer into floating point
        rgba_float32: npt.NDArray[np.float32] = rgba_uint16.astype(np.float32) / (1 << 12)
        return rgba_float32

    return Image(data=Deferred(thunk), dtype=np.float32, channels="rgba")



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from dataclasses import dataclass

from no_vtf.namer import Namer
from no_vtf.textures.vtf import VtfTexture


@dataclass(frozen=True, kw_only=True)
class VTF2TGALikeNamer(Namer[VtfTexture]):
    include_mipmap_level: bool

    def __call__(self, name_stem: str, texture: VtfTexture) -> str:
        if texture.face_index is not None:
            face_names = ("rt", "lf", "bk", "ft", "up", "dn", "sph")
            name_stem += face_names[texture.face_index]

        if texture.frame_index is not None:
            name_stem += f"{texture.frame_index:03d}"

        if self.include_mipmap_level:
            name_stem += f"_mip{texture.mipmap_level}"

        if texture.slice_index is not None:
            name_stem += f"_z{texture.slice_index:03d}"

        return name_stem



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import pathlib
import re

from dataclasses import dataclass
from typing import ClassVar, Optional, cast

import imageio
import imageio.plugins.freeimage
import numpy as np
import numpy.typing as npt

from no_vtf.deferrable import Deferrable, from_deferrable
from no_vtf.image import Image, ImageData, ImageDataTypes
from no_vtf.io import Io
from no_vtf.postprocessor import Postprocessor
from no_vtf.postprocessors.fp_precision_modifier import FloatingPointPrecisionModifier


@dataclass(frozen=True, kw_only=True)
class Imageio(Io[Image[ImageDataTypes]]):

    _format_pattern: ClassVar[re.Pattern[str]] = re.compile(r"[a-z0-9]+", re.ASCII | re.IGNORECASE)

    _freeimage_requested: ClassVar[bool] = False

    @classmethod
    def request_freeimage(cls) -> None:
        if cls._freeimage_requested:
            return

        imageio.plugins.freeimage.download()

        cls._freeimage_requested = True

    format: str

    compress: Optional[bool] = None

    def __post_init__(self) -> None:
        if not self._format_pattern.fullmatch(self.format):
            raise RuntimeError(f"Invalid format: {self.format}")

    def write(self, path: pathlib.Path, image: Deferrable[Image[ImageDataTypes]]) -> None:
        backend = self._get_backend()
        data = self._get_data(image)

        backend.write(path, data)

    def readback(self, path: pathlib.Path, image: Deferrable[Image[ImageDataTypes]]) -> None:
        backend = self._get_backend()
        data = self._get_data(image)

        backend.readback(path, data)

    def _get_backend(self) -> ImageioBackend:
        compress = self.compress
        if compress is None:
            compress = True

        backend: ImageioBackend
        match self.format.lower():
            case "exr":
                backend = ImageioExrBackend(compress=compress)
            case "png":
                backend = ImageioPngBackend(compress=compress)
            case "targa" | "tga":
                backend = ImageioTgaBackend(compress=compress)
            case "tiff":
                backend = ImageioTiffBackend(compress=compress)
            case _:
                backend = ImageioBackend()
        return backend

    def _get_data(self, image: Deferrable[Image[ImageDataTypes]]) -> ImageData:
        image = from_deferrable(image)
        data = image.data()

        # write luminance into three channels when alpha is present
        if image.channels == "la":
            l_uint8: npt.NDArray[np.uint8] = data[:, :, [0]]
            a_uint8: npt.NDArray[np.uint8] = data[:, :, [1]]
            data = np.dstack((l_uint8, l_uint8, l_uint8, a_uint8))

        # remove last axis if its length is 1
        if data.shape[-1] == 1:
            data = data[..., 0]

        return data


class ImageioBackend:
    def __init__(self, *, _imageio_format: Optional[str] = None) -> None:
        self._imageio_format = _imageio_format

    def write(self, path: pathlib.Path, data: ImageData) -> None:
        kwargs = self._get_writer_kwargs(data)
        data = self._postprocess(data)
        with imageio.get_writer(path, format=self._imageio_format, mode="i", **kwargs) as writer:
            writer.append_data(data)

    def readback(self, path: pathlib.Path, data: ImageData) -> None:
        data = self._postprocess(data)
        with imageio.get_reader(path, format=self._imageio_format, mode="i") as reader:
            read_data = reader.get_data(index=0)

            if data.dtype != read_data.dtype:
                raise RuntimeError("Data type differs from what is in the file")

            if not self._compare(data, read_data):
                raise RuntimeError("Data differs from what is in the file")

    def _get_writer_kwargs(self, data: ImageData) -> dict[str, object]:
        return {}

    def _postprocess(self, data: ImageData) -> ImageData:
        return data

    def _compare(self, data: ImageData, read_data: ImageData) -> bool:
        return np.array_equal(data, read_data)


class ImageioFreeImageBackend(ImageioBackend):
    IO_FLAGS: ClassVar = imageio.plugins.freeimage.IO_FLAGS

    def __init__(self, *, _imageio_format: str) -> None:
        super().__init__(_imageio_format=_imageio_format)

    def _get_writer_kwargs(self, data: ImageData) -> dict[str, object]:
        kwargs: dict[str, object] = {}
        kwargs["flags"] = self._get_flags(data)
        return kwargs

    def _get_flags(self, data: ImageData) -> int:
        return 0


class ImageioExrBackend(ImageioFreeImageBackend):
    _fp_force_32_bits: ClassVar[Postprocessor[ImageData]] = FloatingPointPrecisionModifier(
        forced=32
    )

    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(_imageio_format="EXR-FI")
        self.compress = compress

    def _get_flags(self, data: ImageData) -> int:
        flags = 0
        flags |= self.IO_FLAGS.EXR_ZIP if self.compress else self.IO_FLAGS.EXR_NONE
        if not np.issubdtype(data.dtype, np.float16):
            flags |= self.IO_FLAGS.EXR_FLOAT
        return flags

    def _postprocess(self, data: ImageData) -> ImageData:
        return self._fp_force_32_bits(data)


class ImageioPngBackend(ImageioFreeImageBackend):
    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(_imageio_format="PNG-FI")
        self.compress = compress

    def _get_writer_kwargs(self, data: ImageData) -> dict[str, object]:
        kwargs: dict[str, object] = super()._get_writer_kwargs(data)
        kwargs["compression"] = 1 if self.compress else 0
        return kwargs

    def _compare(self, data: ImageData, read_data: ImageData) -> bool:
        if np.issubdtype(data.dtype, np.uint8) and np.issubdtype(read_data.dtype, np.uint8):
            if read_data.ndim == 3 and read_data.shape[2] == 3:
                data = _strip_opaque_alpha(cast(npt.NDArray[np.uint8], data))

        return super()._compare(data, read_data)


class ImageioTgaBackend(ImageioFreeImageBackend):
    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(_imageio_format="TARGA-FI")
        self.compress = compress

    def _get_flags(self, data: ImageData) -> int:
        flags = 0
        flags |= self.IO_FLAGS.TARGA_SAVE_RLE if self.compress else self.IO_FLAGS.TARGA_DEFAULT
        return flags


class ImageioTiffBackend(ImageioFreeImageBackend):
    _fp_force_32_bits: ClassVar[Postprocessor[ImageData]] = FloatingPointPrecisionModifier(
        forced=32
    )

    def __init__(self, *, compress: bool = True) -> None:
        super().__init__(_imageio_format="TIFF-FI")
        self.compress = compress

    def _get_flags(self, data: ImageData) -> int:
        flags = 0
        flags |= self.IO_FLAGS.TIFF_DEFAULT if self.compress else self.IO_FLAGS.TIFF_NONE
        return flags

    def _postprocess(self, data: ImageData) -> ImageData:
        return self._fp_force_32_bits(data)


def _strip_opaque_alpha(data: npt.NDArray[np.uint8]) -> npt.NDArray[np.uint8]:
    if data.ndim == 3 and data.shape[2] == 4 and np.all(data[..., 3] == 255):
        data = data[..., :3]

    return data



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Callable, Generic, Optional, TypeAlias, TypeVar, cast

_DeferrableTypeVar = TypeVar("_DeferrableTypeVar", covariant=True)


class Deferred(Generic[_DeferrableTypeVar]):
    def __init__(self, thunk: Callable[[], _DeferrableTypeVar]) -> None:
        self._thunk = thunk
        self._result: Optional[_DeferrableTypeVar] = None
        self._result_valid = False

    def __call__(self) -> _DeferrableTypeVar:
        if self._result_valid:
            return cast(_DeferrableTypeVar, self._result)

        self._result = self._thunk()
        self._result_valid = True
        return self._result


Deferrable: TypeAlias = _DeferrableTypeVar | Deferred[_DeferrableTypeVar]


def to_deferred(deferrable: Deferrable[_DeferrableTypeVar]) -> Deferred[_DeferrableTypeVar]:
    if isinstance(deferrable, Deferred):
        return deferrable

    not_deferred: _DeferrableTypeVar = deferrable
    return Deferred(lambda: not_deferred)


def from_deferrable(deferrable: Deferrable[_DeferrableTypeVar]) -> _DeferrableTypeVar:
    if isinstance(deferrable, Deferred):
        return deferrable()

    return deferrable



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

import itertools
import os
import pathlib
import shutil
import tempfile

from contextlib import contextmanager
from dataclasses import dataclass
from typing import Iterator, List, Optional, Sequence

import nox

nox.needs_version = ">= 2022.1.7"

nox.options.default_venv_backend = "venv"
nox.options.error_on_external_run = True
nox.options.error_on_missing_interpreters = True
nox.options.sessions = ["lint"]


@nox.session(python="3.10")
def lint(session: nox.Session) -> None:
    session.install("black[colorama] >= 22.3.0, < 23")
    session.install("flake8 >= 4.0.1, < 5")
    session.install("flake8-deprecated >= 1.3, < 2")
    session.install("isort[colors] >= 5.10.1, < 6")
    session.install("mypy >= 0.941, < 1")
    session.install("nox >= 2022.1.7, < 2023")
    session.install("pep8-naming >= 0.12.1, < 1")

    session.install(".")

    posargs_paths = session.posargs
    fix = False
    if posargs_paths and posargs_paths[0] == "--fix":
        posargs_paths = posargs_paths[1:]
        fix = True

    paths = ["no_vtf", "noxfile.py"]
    if posargs_paths:
        paths = posargs_paths

    if not fix:
        session.run("mypy", "--pretty", "--show-error-context", *paths)
        session.run("flake8", *paths)
        session.run("isort", "--check", "--diff", *paths)
        session.run("black", "--check", "--diff", *paths)
    else:
        session.run("isort", *paths)
        session.run("black", *paths)


@nox.session(python="3.10")
def package(session: nox.Session) -> None:
    session.install("build >= 0.7.0, < 1")

    session.run("python", "-m", "build", silent=True)

    path_dist = pathlib.Path("dist")
    path_sdist = next(path_dist.glob("*.tar.gz"))
    path_wheel = next(path_dist.glob("*.whl"))

    # run even with the --no-install flag
    session.run("pip", "install", str(path_wheel), silent=True)

    executable = ["python", "-m", "no_vtf"]
    session.run(*executable, "--version")

    with SanityTest.with_archived_samples(session, executable) as sanity_test:
        sanity_test.readback()

    if len(session.posargs) >= 1:
        shutil.copy2(path_sdist, session.posargs[0])

    if len(session.posargs) >= 2:
        shutil.copy2(path_wheel, session.posargs[1])


@nox.session(python="3.10")
def freeze(session: nox.Session) -> None:
    session.install("pyinstaller >= 4.9, < 5")

    # installing in editable mode will ensure the _version.py file is generated by setuptools_scm
    # run even with the --no-install flag
    session.run("pip", "install", "-e", ".", silent=True)

    session.run("imageio_download_bin", "--package-dir", silent=True)

    pyinstaller_datas: list[tuple[str, str]] = []
    pyinstaller_args: list[str] = []

    pyinstaller_args.extend(["--log-level", "WARN"])
    pyinstaller_args.extend(["--name", "no_vtf"])
    pyinstaller_args.extend(["--icon", "resources/pyinstaller/empty.ico"])

    if os.name == "posix":
        pyinstaller_datas.append(("resources/pyinstaller/no_vtf.desktop", "."))

    # https://github.com/rsalmei/alive-progress/issues/123
    pyinstaller_args.extend(["--collect-data", "grapheme"])

    # https://github.com/imageio/imageio/issues/690#issuecomment-974605991
    # plugin list: imageio/docs/conf.py
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.bsdf"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.dicom"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.feisem"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.ffmpeg"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.fits"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.freeimage"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.gdal"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.lytro"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.npz"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.pillow_legacy"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.simpleitk"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.spe"])
    pyinstaller_args.extend(["--hidden-import", "imageio.plugins.swf"])

    # bundle plugin binary dependencies downloaded by imageio_download_bin
    pyinstaller_args.extend(["--collect-binaries", "imageio"])

    pyinstaller_args.append("--noconfirm")
    pyinstaller_args.append("--clean")

    for pyinstaller_data in pyinstaller_datas:
        pyinstaller_args.extend(["--add-data", os.pathsep.join(pyinstaller_data)])

    session.run("pyinstaller", *pyinstaller_args, "no_vtf/__main__.py")
    shutil.rmtree("dist/no_vtf/imageio/resources/images", ignore_errors=True)

    executable = "dist/no_vtf/no_vtf"
    paths = ["."]
    nox.command.run([executable, "--version"], paths=paths, external=True)

    with SanityTest.with_archived_samples(session, [executable], paths) as sanity_test:
        sanity_test.readback()

    if not session.posargs:
        return

    archive_path = session.posargs[0]
    archive_base_name: str
    archive_format: str
    if (archive_base_name := archive_path.removesuffix(".tar.xz")) != archive_path:
        archive_format = "xztar"
    elif (archive_base_name := archive_path.removesuffix(".zip")) != archive_path:
        archive_format = "zip"
    else:
        raise RuntimeError("Unsupported archive format")

    shutil.make_archive(archive_base_name, archive_format, root_dir="dist", base_dir="no_vtf")


@nox.session(python="3.10")
def publish(session: nox.Session) -> None:
    if not session.posargs:
        session.error("Path to API token file was not provided")

    session.install("twine >= 4.0.0, < 5")

    dist = pathlib.Path("dist")
    dist_files = [path for path in dist.iterdir() if path.is_file()]
    dist_args = [str(path) for path in dist_files]

    session.run("twine", "check", "--strict", *dist_args)

    upload_args = []
    upload_args.append("--non-interactive")
    upload_args.append("--disable-progress-bar")
    upload_args.extend(dist_args)

    env = session.env.copy()
    env["TWINE_USERNAME"] = "__token__"
    env["TWINE_PASSWORD"] = pathlib.Path(session.posargs[0]).read_text().strip()

    nox.command.run(
        ["twine", "upload", *upload_args],
        env=env,
        paths=session.bin_paths,
        external=True,
    )


@nox.session(python="3.10")
def write_reference(session: nox.Session) -> None:
    if not session.posargs:
        session.error("Path to test samples was not provided")

    # run even with the --no-install flag
    session.run("pip", "install", ".", silent=True)

    executable = ["python", "-m", "no_vtf"]

    base_path = pathlib.Path(session.posargs[0])
    input = base_path / "input"
    output = base_path / "output"

    sanity_test = SanityTest(session=session, executable=executable, input=input, output=output)
    sanity_test.write()


@dataclass(frozen=True, kw_only=True)
class SanityTest:
    session: nox.Session
    executable: Sequence[str]
    paths: Optional[List[str]] = None

    input: pathlib.Path
    output: pathlib.Path

    @classmethod
    @contextmanager
    def with_archived_samples(
        cls, session: nox.Session, executable: Sequence[str], paths: Optional[List[str]] = None
    ) -> Iterator[SanityTest]:
        with samples_archive_view() as samples_directory:
            input = samples_directory / "input"
            output = samples_directory / "output"
            sanity_test = cls(
                session=session, executable=executable, paths=paths, input=input, output=output
            )
            yield sanity_test

    def write(self) -> None:
        args: list[str] = []
        args.append("--always-write")
        args.append("--no-compress")

        self._run_combinations(args)

    def readback(self) -> None:
        args: list[str] = []
        args.append("--no-write")
        args.append("--readback")

        self._run_combinations(args)

    def _run_combinations(self, args: Sequence[str]) -> None:
        mipmaps = [[], ["--mipmaps"]]
        separate_channels = [[], ["--separate-channels"]]

        combinations = [mipmaps, separate_channels]
        for combination in itertools.product(*combinations):
            combination_flattened = itertools.chain.from_iterable(combination)
            self._run_formats([*args, *combination_flattened])

    def _run_formats(self, args: Sequence[str]) -> None:
        args = list(args)

        ldr_formats = ["raw", "png", "tga", "tiff"]
        for ldr_format in ldr_formats:
            self._run([*args, "--ldr-format", ldr_format, "--hdr-format", "skip"])

        hdr_formats = ["raw", "exr", "tiff"]
        for hdr_format in hdr_formats:
            self._run([*args, "--ldr-format", "skip", "--hdr-format", hdr_format])

    def _run(self, args: Sequence[str]) -> None:
        args = list(args)

        args.append("--no-progress")

        args.append(str(self.input))
        args.extend(["--output-dir", str(self.output)])

        paths = self.paths or self.session.bin_paths
        nox.command.run([*self.executable, *args], env=self.session.env, paths=paths, external=True)


@contextmanager
def archive_view(archive: pathlib.Path) -> Iterator[pathlib.Path]:
    with archives_view([archive]) as view_directory:
        yield view_directory


@contextmanager
def archives_view(archives: Sequence[pathlib.Path]) -> Iterator[pathlib.Path]:
    with tempfile.TemporaryDirectory() as view_directory:
        for archive in archives:
            shutil.unpack_archive(archive, view_directory)
        yield pathlib.Path(view_directory)


@contextmanager
def samples_archive_view() -> Iterator[pathlib.Path]:
    samples_path = pathlib.Path("resources/test/samples")
    sample_archives = [path for path in samples_path.iterdir() if path.is_file()]
    with archives_view(sample_archives) as samples_directory:
        yield samples_directory



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Iterable, Optional, TypeVar

import numpy.typing as npt

from no_vtf.deferrable import Deferrable, Deferred, from_deferrable
from no_vtf.image import Image, ImageDataTypes

_ImageDataTypeVar = TypeVar(
    "_ImageDataTypeVar",
    bound=ImageDataTypes,
    contravariant=True,
)


class ChannelSeparator:
    def __call__(
        self, image: Deferrable[Image[_ImageDataTypeVar]]
    ) -> Iterable[Image[_ImageDataTypeVar]]:
        image = from_deferrable(image)

        data_rgb: Optional[Deferred[npt.NDArray[_ImageDataTypeVar]]] = None
        data_l: Optional[Deferred[npt.NDArray[_ImageDataTypeVar]]] = None
        data_a: Optional[Deferred[npt.NDArray[_ImageDataTypeVar]]] = None

        match (image.channels):
            case "rgb":
                data_rgb = image.data
            case "rgba":
                data_rgb = _index_channels(image.data, slice(0, 3))
                data_a = _index_channels(image.data, slice(3, 4))
            case "l":
                data_l = image.data
            case "la":
                data_l = _index_channels(image.data, slice(0, 1))
                data_a = _index_channels(image.data, slice(1, 2))
            case "a":
                data_a = image.data
            case _:
                raise RuntimeError(f"Unsupported channel configuration: {image.channels}")

        if data_rgb is not None:
            yield Image(data=data_rgb, dtype=image.dtype, channels="rgb")

        if data_l is not None:
            yield Image(data=data_l, dtype=image.dtype, channels="l")

        if data_a is not None:
            yield Image(data=data_a, dtype=image.dtype, channels="a")


def _index_channels(
    data: Deferrable[npt.NDArray[_ImageDataTypeVar]], channels: slice
) -> Deferred[npt.NDArray[_ImageDataTypeVar]]:
    def thunk() -> npt.NDArray[_ImageDataTypeVar]:
        _data = from_deferrable(data)
        _data = _data[..., channels]
        return _data

    return Deferred(thunk)



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from __future__ import annotations

import collections
import pathlib

from dataclasses import dataclass
from typing import Iterable, Optional, Sequence


class DirectoryFiles:
    def __init__(self, path: pathlib.Path) -> None:
        self.directory: pathlib.Path
        self.files: list[pathlib.Path]

        if path.is_dir():
            self.directory = path
            self.files = []
        else:
            self.directory = path.parent
            self.files = [path]

    def search_in_directory(self, pattern: str) -> Iterable[pathlib.Path]:
        assert not self.files

        for file in self.directory.rglob(pattern):
            self.files.append(file)
            yield file


class DirectoryFilesList(collections.UserList[DirectoryFiles]):
    @staticmethod
    def from_paths(paths: Sequence[pathlib.Path]) -> DirectoryFilesList:
        directory_files_list = DirectoryFilesList()
        for path in paths:
            directory_files = DirectoryFiles(path)
            directory_files_list.append(directory_files)
        return directory_files_list

    def __init__(self, initlist: Optional[Iterable[DirectoryFiles]] = None):
        super().__init__(initlist)

    def num_files(self) -> int:
        num_files = 0
        for files in self._files_list():
            num_files += len(files)
        return num_files

    def has_unresolved_files(self) -> bool:
        return not all(self._files_list())

    def search_in_directories(self, pattern: str) -> Iterable[pathlib.Path]:
        for directory_files in self.data:
            if not directory_files.files:
                for file in directory_files.search_in_directory(pattern):
                    yield file

    def _files_list(self) -> Sequence[Sequence[pathlib.Path]]:
        return list(map(lambda directory_files: directory_files.files, self.data))


@dataclass(frozen=True, kw_only=True)
class FileAccessor:
    skip_existing: bool = False
    mkdir_missing_parents: bool = False

    def __call__(self, path: pathlib.Path) -> bool:
        if self.skip_existing and path.is_file():
            return False

        if self.mkdir_missing_parents:
            path.parent.mkdir(parents=True, exist_ok=True)

        return True



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

import typing

from dataclasses import dataclass
from typing import Generic, Literal, TypeAlias, TypeVar, Union

import numpy as np
import numpy.typing as npt

from no_vtf.deferrable import Deferred

DynamicRange = Literal["ldr", "hdr"]

ImageDataTypesLDR: TypeAlias = np.uint8 | np.uint16
ImageDataTypesHDR: TypeAlias = np.float16 | np.float32

ImageDataTypes: TypeAlias = Union[ImageDataTypesLDR, ImageDataTypesHDR]

ImageData: TypeAlias = npt.NDArray[ImageDataTypes]

_ImageDataTypeVar = TypeVar(
    "_ImageDataTypeVar",
    bound=ImageDataTypes,
    covariant=True,
)

ImageChannels = Literal["rgb", "rgba", "l", "la", "a"]


@dataclass(frozen=True, kw_only=True)
class Image(Generic[_ImageDataTypeVar]):
    data: Deferred[npt.NDArray[_ImageDataTypeVar]]
    dtype: npt.DTypeLike
    channels: ImageChannels

    def get_dynamic_range(self) -> DynamicRange:
        ldr = is_ldr(self.dtype)
        hdr = is_hdr(self.dtype)
        assert ldr != hdr

        return "hdr" if hdr else "ldr"


def is_ldr(dtype: npt.DTypeLike) -> bool:
    ldr_dtypes = typing.get_args(ImageDataTypesLDR)
    for ldr_dtype in ldr_dtypes:
        if np.issubdtype(dtype, ldr_dtype):
            return True

    return False


def is_hdr(dtype: npt.DTypeLike) -> bool:
    hdr_dtypes = typing.get_args(ImageDataTypesHDR)
    for hdr_dtype in hdr_dtypes:
        if np.issubdtype(dtype, hdr_dtype):
            return True

    return False



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from typing import Iterable, Optional, Sequence, TypeVar

from no_vtf.pipeline import Pipeline
from no_vtf.pipeline_runner import PipelineRunner, Result, Task
from no_vtf.texture import Texture

_TextureTypeVar = TypeVar("_TextureTypeVar", bound=Texture)


class SequentialRunner(PipelineRunner):
    def __call__(
        self, pipeline: Pipeline[_TextureTypeVar], tasks: Sequence[Task]
    ) -> Iterable[Result]:
        for task in tasks:
            yield self.process(pipeline, task)

    @staticmethod
    def process(pipeline: Pipeline[_TextureTypeVar], task: Task) -> Result:
        exception: Optional[Exception] = None
        try:
            pipeline(task.input_file, task.output_directory)
        except Exception as e:
            exception = e

        return Result(task=task, exception=exception)



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod


class Texture:
    @abstractmethod
    def get_width(self) -> int:
        ...

    @abstractmethod
    def get_height(self) -> int:
        ...

    @abstractmethod
    def get_data(self) -> bytes:
        ...



# Copyright (C) 2022 b5327157
# SPDX-License-Identifier: LGPL-3.0-or-later

from abc import abstractmethod
from typing import Protocol, Sequence, TypeAlias, TypeVar

from no_vtf.image import Image, ImageDataTypes

_PostprocessorTypeVar = TypeVar("_PostprocessorTypeVar")


class Postprocessor(Protocol[_PostprocessorTypeVar]):
    # The argument is declared as positional-only, so implementations can rename it.
    # Warning: mypy does not check argument name in implementation, even if not declared as such.
    @abstractmethod
    def __call__(self, data: _PostprocessorTypeVar, /) -> _PostprocessorTypeVar:
        ...


def apply_postprocessors(
    postprocessors: Sequence[Postprocessor[_PostprocessorTypeVar]], data: _PostprocessorTypeVar
) -> _PostprocessorTypeVar:
    for postprocessor in postprocessors:
        data = postprocessor(data)
    return data


_ImageDataTypeVar = TypeVar("_ImageDataTypeVar", bound=ImageDataTypes, covariant=True)

ImagePostprocessor: TypeAlias = Postprocessor[Image[_ImageDataTypeVar]]
